@@ -1,11 +1,17 @@
-from typing import Any, Dict
+from typing import TYPE_CHECKING, Any, Dict
 
 from novelai_api.BiasGroup import BiasGroup
 from novelai_api.Preset import Model
+from novelai_api.python_utils import expand_kwargs
 from novelai_api.Tokenizer import Tokenizer
 
 
 class GlobalSettings:
+    """
+    Object used to store global settings for the account
+    """
+
+    # TODO: store bracket ban in a file
     _BRACKETS = {
         "gpt2": [
             [58],
@@ -463,6 +469,23 @@ class GlobalSettings:
             [0],
             [1],
         ],
+        "nerdstash_v1": [
+            [3],
+            [49356],
+            [1431],
+            [31715],
+            [34387],
+            [20765],
+            [30702],
+            [10691],
+            [49333],
+            [1266],
+            [19438],
+            [43145],
+            [26523],
+            [41471],
+            [2936],
+        ],
     }
 
     # this one is pretty much mandatory, else genji throws errors all the time
@@ -594,32 +617,73 @@ class GlobalSettings:
         "ban_ambiguous_genji_tokens": True,
     }
 
+    # type completion for __setitem__ and __getitem__
+    if TYPE_CHECKING:
+        #: Generate up to 20 tokens after max_length if an end of sentence if found within these 20 tokens
+        generate_until_sentence: bool
+        #: Number of logprobs to return for each token. Set to NO_LOGPROBS to disable
+        num_logprobs: int
+        #: Apply the BRACKET biases
+        ban_brackets: bool
+        #: Apply the DINKUS_ASTERISM biases
+        bias_dinkus_asterism: bool
+        #: Apply the GENJI_AMBIGUOUS_TOKENS if model is Genji
+        ban_ambiguous_genji_tokens: bool
+
+    #: Value to set num_logprobs at to disable logprobs
     NO_LOGPROBS = -1
 
     _settings: Dict[str, Any]
 
+    @expand_kwargs(_DEFAULT_SETTINGS.keys(), (type(e) for e in _DEFAULT_SETTINGS.values()))
     def __init__(self, **kwargs):
-        self._settings = {}
+        object.__setattr__(self, "_settings", {})
 
         for setting, default in self._DEFAULT_SETTINGS.items():
             self._settings[setting] = kwargs.pop(setting, default)
 
-        assert len(kwargs) == 0, f"Invalid global setting name: {', '.join(kwargs)}"
+        if kwargs:
+            raise ValueError(f"Invalid global setting name: {', '.join(kwargs)}")
 
-    def __setitem__(self, o: str, v: Any) -> None:
-        assert o in self._settings, f"Invalid setting: {o}"
+    def __setitem__(self, key: str, value: Any) -> None:
+        if key not in self._DEFAULT_SETTINGS:
+            raise ValueError(f"Invalid setting: '{key}'")
 
-        self._settings[o] = v
+        self._settings[key] = value
 
-    def __getitem__(self, o: str) -> Any:
-        assert o in self._settings, f"Invalid setting: {o}"
+    def __getitem__(self, key: str) -> Any:
+        if key not in self._DEFAULT_SETTINGS:
+            raise ValueError(f"Invalid setting: '{key}'")
 
-        return self._settings[o]
+        return self._settings[key]
+
+    # give dot access capabilities to the object
+    def __setattr__(self, key, value):
+        if key in self._DEFAULT_SETTINGS:
+            self[key] = value
+        else:
+            object.__setattr__(self, key, value)
+
+    def __getattr__(self, key):
+        if key in self._DEFAULT_SETTINGS:
+            return self[key]
+
+        return object.__getattribute__(self, key)
 
     def copy(self):
+        """
+        Create a new GlobalSettings from the current
+        """
+
         return GlobalSettings(**self._settings)
 
     def to_settings(self, model: Model) -> Dict[str, Any]:
+        """
+        Create text generation settings from the GlobalSettings object
+
+        :param model: Model to use the settings of
+        """
+
         settings = {
             "generate_until_sentence": self._settings["generate_until_sentence"],
             "num_logprobs": self._settings["num_logprobs"],
@@ -630,6 +694,10 @@ class GlobalSettings:
             "use_cache": False,
         }
 
+        # NO_LOGPROBS is used to disable logprobs (=> not in the call)
+        if self._settings["num_logprobs"] == self.NO_LOGPROBS:
+            del settings["num_logprobs"]
+
         tokenizer_name = Tokenizer.get_tokenizer_name(model)
 
         if self._settings["ban_brackets"]:
@@ -639,6 +707,6 @@ class GlobalSettings:
             settings["bad_words_ids"].extend(self._GENJI_AMBIGUOUS_TOKENS)
 
         if self._settings["bias_dinkus_asterism"]:
-            settings["logit_bias_exp"].extend(self._DINKUS_ASTERISM.get_tokenized_biases(model))
+            settings["logit_bias_exp"].extend(self._DINKUS_ASTERISM.get_tokenized_entries(model))
 
         return settings
