@@ -1,18 +1,30 @@
-import base64
+import json
 from hashlib import sha256
-from typing import Any, AsyncIterable, Dict, Iterable, List, Optional, Tuple, Union
+from typing import Any, AsyncIterable, Dict, Iterable, List, Optional, Tuple, Type, Union
 
 from novelai_api.BanList import BanList
 from novelai_api.BiasGroup import BiasGroup
 from novelai_api.GlobalSettings import GlobalSettings
-from novelai_api.ImagePreset import ImageModel, ImagePreset
+from novelai_api.ImagePreset import ImageGenerationType, ImageModel, ImagePreset
 from novelai_api.Keystore import Keystore
 from novelai_api.NovelAIError import NovelAIError
 from novelai_api.Preset import Model, Preset
-from novelai_api.utils import assert_type, compress_user_data, encrypt_user_data, get_access_key
+from novelai_api.python_utils import assert_type
+from novelai_api.utils import compress_user_data, encrypt_user_data, get_access_key
 
 
 class HighLevel:
+    """
+    High level API for NovelAI. This class is not meant to be used directly,
+    but rather through :attr:`NovelAIAPI.high_level <novelai_api.NovelAI_API.NovelAIAPI.high_level>`.
+
+    The most relevant methods are:
+
+    * :meth:`login <novelai_api._high_level.HighLevel.login>`
+    * :meth:`generate <novelai_api._high_level.HighLevel.generate>`
+    * :meth:`generate_image <novelai_api._high_level.HighLevel.generate_image>`
+    """
+
     _parent: "NovelAIAPI"  # noqa: F821
 
     def __init__(self, parent: "NovelAIAPI"):  # noqa: F821
@@ -38,7 +50,7 @@ class HighLevel:
         :return: True if success
         """
 
-        assert assert_type(str, email=email)
+        assert_type(str, email=email)
 
         hashed_email = sha256(email.encode()).hexdigest() if send_mail else None
         key = get_access_key(email, password)
@@ -62,13 +74,24 @@ class HighLevel:
         return rsp["accessToken"]
 
     async def login_from_token(self, access_key: str):
+        """
+        Log in with the access key, instead of email and password
+
+        :param access_key: Access key of the account (pre-computed via email and password)
+
+        :return: User's access token
+        """
+
         rsp = await self._parent.low_level.login(access_key)
 
         self._parent.headers["Authorization"] = f"Bearer {rsp['accessToken']}"
 
+        return rsp["accessToken"]
+
     async def get_keystore(self, key: bytes) -> Keystore:
         """
         Retrieve the keystore and decrypt it in a readable manner.
+
         The keystore is the mapping of meta -> encryption key of each object.
         If this function throws errors repeatedly at you,
         check your internet connection or the integrity of your keystore.
@@ -85,33 +108,67 @@ class HighLevel:
         return keystore
 
     async def set_keystore(self, keystore: Keystore, key: bytes) -> bytes:
+        """
+        Encrypt and upload the keystore.
+
+        The keystore is the mapping of meta -> encryption key of each object.
+        If this function throws errors repeatedly at you,
+        check your internet connection or the integrity of your keystore.
+        Losing your keystore, or overwriting it means losing all content on the account.
+
+        :param keystore: Keystore object to upload
+        :param key: Account's encryption key
+
+        :return: raw data of the serialized Keystore object
+        """
+
         keystore.encrypt(key)
 
         return await self._parent.low_level.set_keystore(keystore.data)
 
-    async def download_user_stories(self) -> Dict[str, Dict[str, Union[str, int]]]:
+    async def download_user_stories(self) -> List[Dict[str, Dict[str, Union[str, int]]]]:
+        """
+        Download all the objects of type 'stories' stored on the account
+        """
+
         stories = await self._parent.low_level.download_objects("stories")
 
         return stories["objects"]
 
     async def download_user_story_contents(
         self,
-    ) -> Dict[str, Dict[str, Union[str, int]]]:
+    ) -> List[Dict[str, Dict[str, Union[str, int]]]]:
+        """
+        Download all the objects of type 'storycontent' stored on the account
+        """
+
         story_contents = await self._parent.low_level.download_objects("storycontent")
 
         return story_contents["objects"]
 
     async def download_user_presets(self) -> List[Dict[str, Union[str, int]]]:
+        """
+        Download all the objects of type 'presets' stored on the account
+        """
+
         presets = await self._parent.low_level.download_objects("presets")
 
         return presets["objects"]
 
     async def download_user_modules(self) -> List[Dict[str, Union[str, int]]]:
+        """
+        Download all the objects of type 'aimodules' stored on the account
+        """
+
         modules = await self._parent.low_level.download_objects("aimodules")
 
         return modules["objects"]
 
     async def download_user_shelves(self) -> List[Dict[str, Union[str, int]]]:
+        """
+        Download all the objects of type 'shelf' stored on the account
+        """
+
         modules = await self._parent.low_level.download_objects("shelf")
 
         return modules["objects"]
@@ -123,12 +180,11 @@ class HighLevel:
         keystore: Optional[Keystore] = None,
     ) -> bool:
         """
-        Upload user content. If it has been decrypted with decrypt_user_data,
-        it should be re-encrypted with encrypt_user_data, even if the decryption failed
+        Upload user content
 
         :param data: Object to upload
-        :param encrypt: Encrypt/compress the data if True and not already encrypted
-        :param keystore: Keystore to encrypt data if encrypt is True
+        :param encrypt: Re-encrypt/re-compress the data, if True
+        :param keystore: Keystore to encrypt the data, if encrypt is True
 
         :return: True if the upload succeeded, False otherwise
         """
@@ -140,7 +196,9 @@ class HighLevel:
 
         if encrypt:
             if object_type in ("stories", "storycontent", "aimodules", "shelf"):
-                assert keystore is not None, "Keystore is not set, cannot encrypt data"
+                if keystore is None:
+                    raise ValueError("'keystore' is not set, cannot encrypt data")
+
                 encrypt_user_data(data, keystore)
             elif object_type in ("presets",):
                 compress_user_data(data)
@@ -154,12 +212,19 @@ class HighLevel:
 
         return await self._parent.low_level.upload_object(object_type, object_id, object_meta, object_data)
 
-    async def upload_user_contents(self, datas: Iterable[Dict[str, Any]]) -> List[Tuple[str, Optional[NovelAIError]]]:
+    async def upload_user_contents(
+        self,
+        datas: Iterable[Dict[str, Any]],
+        encrypt: bool = False,
+        keystore: Optional[Keystore] = None,
+    ) -> List[Tuple[str, Optional[NovelAIError]]]:
         """
         Upload multiple user contents. If the content has been decrypted with decrypt_user_data,
         it should be re-encrypted with encrypt_user_data, even if the decryption failed
 
         :param datas: Objects to upload
+        :param encrypt: Re-encrypt/re-compress the data, if True
+        :param keystore: Keystore to encrypt the data, if encrypt is True
 
         :return: A list of (id, error) of all the objects that failed to be uploaded
         """
@@ -168,7 +233,8 @@ class HighLevel:
 
         for data in datas:
             try:
-                success = await self.upload_user_content(data)
+                success = await self.upload_user_content(data, encrypt, keystore)
+
                 if not success:
                     status.append((data["id"], None))
             except NovelAIError as e:
@@ -189,7 +255,7 @@ class HighLevel:
         **kwargs,
     ):
         """
-        Generate content from an AI on the NovelAI server which support streaming
+        Generate text with streaming support
 
         :param prompt: Context to give to the AI (raw text or list of tokens)
         :param model: Model to use for the AI
@@ -199,14 +265,15 @@ class HighLevel:
         :param biases: Tokens to bias (up or down) for this generation
         :param prefix: Module to use for this generation
         :param stream: Use data streaming for the response
+        :param kwargs: Additional parameters to pass to the requests. Can also be used to overwrite existing parameters
 
         :return: Content that has been generated
         """
 
-        assert preset is not None, "Uninitialized preset"
-        assert (
-            preset.model == model
-        ), f"Preset {preset.name} (model {preset.model}) is not compatible with model {model}"
+        if preset is None:
+            raise ValueError("Uninitialized preset")
+        if preset.model is not model:
+            raise ValueError(f"Preset '{preset.name}' (model {preset.model}) is not compatible with model {model}")
 
         preset_params = preset.to_settings()
         global_params = global_settings.to_settings(model)
@@ -217,37 +284,30 @@ class HighLevel:
         params.update(global_params)
         params.update(kwargs)
 
+        # adjust repetition penalty value for Sigurd and Euterpe
+        if model in (Model.Sigurd, Model.Euterpe) and "repetition_penalty" in params:
+            rep_pen = params["repetition_penalty"]
+            params["repetition_penalty"] = (0.525 * (rep_pen - 1) / 7) + 1
+
         params["prefix"] = "vanilla" if prefix is None else prefix
 
-        if params["num_logprobs"] == GlobalSettings.NO_LOGPROBS:
-            del params["num_logprobs"]
+        for k, v, c in (("bad_words_ids", bad_words, BanList), ("logit_bias_exp", biases, BiasGroup)):
+            k: str
+            v: Union[Iterable[BanList], Iterable[BiasGroup], BanList, BiasGroup, None]
+            c: Union[Type[BanList], Type[BiasGroup]]
 
-        if bad_words is not None:
-            if isinstance(bad_words, BanList):
-                bad_words = [bad_words]
+            if v is not None:
+                if isinstance(v, c):
+                    v = [v]
 
-            for i, bad_word in enumerate(bad_words):
-                assert isinstance(
-                    bad_word, BanList
-                ), f"Expected type 'BanList' for item #{i} of bad_words, but got '{type(bad_word)}'"
-                params["bad_words_ids"].extend(bad_word.get_tokenized_banlist(model))
+                for i, obj in enumerate(v):
+                    if not isinstance(obj, c):
+                        raise ValueError(f"Expected type '{c}' for item #{i} of '{k}', but got '{type(obj)}'")
 
-        if biases is not None:
-            if isinstance(biases, BiasGroup):
-                biases = [biases]
+                    params[k].extend(obj.get_tokenized_entries(model))
 
-            for i, bias in enumerate(biases):
-                assert isinstance(
-                    bias, BiasGroup
-                ), f"Expected type 'BiasGroup' for item #{i} of biases, but got '{type(bias)}'"
-                params["logit_bias_exp"].extend(bias.get_tokenized_biases(model))
-
-        # Delete the options that return an unknown error (success status code, but server error)
-        if "repetition_penalty_slope" in params and params["repetition_penalty_slope"] == 0:
-            del params["repetition_penalty_slope"]
-
-        if not params["bad_words_ids"]:
-            del params["bad_words_ids"]
+            if k in params and not params[k]:
+                del params[k]
 
         async for i in self._parent.low_level.generate(prompt, model, params, stream):
             yield i
@@ -264,7 +324,9 @@ class HighLevel:
         **kwargs,
     ) -> Dict[str, Any]:
         """
-        Generate text from an AI on the NovelAI server which support streaming
+        Generate text. The b64-encoded text is returned at once, when generation is finished.
+        To decode the text, the :meth:`novelai_api.utils.b64_to_tokens`
+        and :meth:`novelai_api.Tokenizer.Tokenizer.decode` methods should be used.
 
         :param prompt: Context to give to the AI (raw text or list of tokens)
         :param model: Model to use for the AI
@@ -273,6 +335,7 @@ class HighLevel:
         :param bad_words: Tokens to ban for this generation
         :param biases: Tokens to bias (up or down) for this generation
         :param prefix: Module to use for this generation
+        :param kwargs: Additional parameters to pass to the requests. Can also be used to overwrite existing parameters
 
         :return: Content that has been generated
         """
@@ -302,7 +365,7 @@ class HighLevel:
         **kwargs,
     ) -> AsyncIterable[Dict[str, Any]]:
         """
-        Generate text from an AI on the NovelAI server
+        Generate text. The text is returned one token at a time, as it is generated.
 
         :param prompt: Context to give to the AI (raw text or list of tokens)
         :param model: Model to use for the AI
@@ -311,6 +374,7 @@ class HighLevel:
         :param bad_words: Tokens to ban for this generation
         :param biases: Tokens to bias (up or down) for this generation
         :param prefix: Module to use for this generation
+        :param kwargs: Additional parameters to pass to the requests. Can also be used to overwrite existing parameters
 
         :return: Content that has been generated
         """
@@ -326,35 +390,39 @@ class HighLevel:
             True,
             **kwargs,
         ):
-            yield e
+            yield json.loads(e)
 
     async def generate_image(
-        self, prompt: str, model: ImageModel, preset: ImagePreset, **kwargs
-    ) -> AsyncIterable[bytes]:
+        self,
+        prompt: str,
+        model: ImageModel,
+        preset: ImagePreset,
+        action: ImageGenerationType = ImageGenerationType.NORMAL,
+        **kwargs,
+    ) -> AsyncIterable[Union[str, bytes]]:
         """
-        Generate image from an AI on the NovelAI server
+        Generate one or multiple image(s)
 
         :param prompt: Prompt to give to the AI (raw text describing the wanted image)
         :param model: Model to use for the AI
         :param preset: Preset to use for the generation settings
+        :param action: Type of image generation to use
+        :param kwargs: Additional parameters to pass to the requests. Can also be used to overwrite existing parameters
 
-        :return: Content that has been generated
+        :return: Pair(s) (name, image) that have been generated
         """
 
         settings = preset.to_settings(model)
         settings.update(kwargs)
 
-        uc = settings["uc"]
+        # TODO: allow to disable the removal of nsfw ? How to tackle this cleanly
+        uc = settings["negative_prompt"]
         if "nsfw" in prompt and uc.startswith("nsfw,"):
-            settings["uc"] = uc[len("nsfw,") :].strip()
+            settings["negative_prompt"] = uc[len("nsfw,") :].strip()
 
-        quality_toggle = preset["quality_toggle"]
+        quality_toggle = settings["qualityToggle"]
         if quality_toggle:
             prompt = f"masterpiece, best quality, {prompt}"
 
-        async for e in self._parent.low_level.generate_image(prompt, model, settings):
-            # FIXME: check back when normalized
-            if "api2" in self._parent.BASE_ADDRESS:
-                yield e
-            else:
-                yield base64.b64decode(e)
+        async for e in self._parent.low_level.generate_image(prompt, model, action, settings):
+            yield e
