@@ -1,5 +1,4 @@
 import json
-import operator
 from base64 import b64decode, b64encode, urlsafe_b64encode
 from hashlib import blake2b
 from typing import Any, Dict, Iterable, List, Optional, Tuple, Union
@@ -12,33 +11,11 @@ from nacl.exceptions import CryptoError
 from nacl.secret import SecretBox
 
 from novelai_api.Keystore import Keystore
+from novelai_api.NovelAIError import NovelAIError
 from novelai_api.Preset import Model, Preset
+from novelai_api.python_utils import assert_type
 from novelai_api.Tokenizer import Tokenizer
 
-# boilerplate
-NoneType: type = type(None)
-
-
-def assert_type(expected, **types):
-    for k, v in types.items():
-        assert isinstance(v, expected), f"Expected type '{expected}' for {k}, but got type '{type(v)}'"
-
-
-operator_to_str = {
-    operator.eq: "exactly {} characters",
-    operator.lt: "less than {} characters",
-    operator.le: "{} characters or less",
-    operator.gt: "more than {} characters",
-    operator.ge: "{} characters or more",
-}
-
-
-def assert_len(expected, op: operator = operator.eq, **values):
-    op_str = operator_to_str[op].format(expected)
-
-    for k, v in values.items():
-        assert v is None or op(len(v), expected), f"'{k}' should be {op_str}, got length of {len(v)}'"
-
 
 # API utils
 def argon_hash(email: str, password: str, size: int, domain: str) -> str:
@@ -151,9 +128,11 @@ def decompress_user_data(items: Union[List[Dict[str, Any]], Dict[str, Any]]):
     if not isinstance(items, (list, tuple)):
         items = [items]
 
-    for item in items:
-        assert isinstance(item, dict), f"Expected type 'dict' for item of 'items', got type '{type(item)}'"
-        assert "data" in item, "Expected key 'data' in item"
+    for i, item in enumerate(items):
+        if not isinstance(item, dict):
+            raise ValueError(f"Expected type 'dict' for item #{i} of 'items', got type '{type(item)}'")
+        if "data" not in item:
+            raise ValueError(f"Expected key 'data' in item #{i} of 'items'")
 
         # skip already decompressed data
         if item.get("decrypted"):
@@ -184,9 +163,12 @@ def compress_user_data(items: Union[List[Dict[str, Any]], Dict[str, Any]]):
     if not isinstance(items, (list, tuple)):
         items = [items]
 
-    for item in items:
-        assert isinstance(item, dict), f"Expected type 'dict' for item of 'items', got type '{type(item)}'"
-        assert "data" in item, "Expected key 'data' in item"
+    for i, item in enumerate(items):
+        if not isinstance(item, dict):
+            raise ValueError(f"Expected type 'dict' for item #{i} of 'items', got type '{type(item)}'")
+
+        if "data" not in item:
+            raise ValueError(f"Expected key 'data' in item #{i} of 'items'")
 
         if "decrypted" in item:
             if item["decrypted"]:
@@ -216,35 +198,36 @@ def decrypt_user_data(items: Union[List[Dict[str, Any]], Dict[str, Any]], keysto
     if not isinstance(items, (list, tuple)):
         items = [items]
 
-    for item in items:
-        assert isinstance(item, dict), f"Expected type 'dict' for item of 'items', got type '{type(item)}'"
+    for i, item in enumerate(items):
+        if not isinstance(item, dict):
+            raise ValueError(f"Expected type 'dict' for item #{i} of 'items', got type '{type(item)}'")
 
         if item.get("decrypted"):
             continue
 
-        # FIXME: replace the assert by meaningful errors ? Expect the data to be right ?
-        assert "data" in item, "Expected key 'data' in item"
-        assert "meta" in item, "Expected key 'meta' in item"
+        if "data" not in item:
+            raise ValueError(f"Expected key 'data' in item #{i} of 'items'")
+        if "meta" not in item:
+            raise ValueError(f"Expected key 'meta' in item #{i} of 'items'")
 
         meta = item["meta"]
-        #       assert meta in keystore
         if meta not in keystore:
-            print("Meta missing:", meta)
-        else:
-            key = keystore[meta]
+            raise NovelAIError("<UNKNOWN>", -1, f"Meta of item #{i} ({meta}) missing from keystore")
 
-            data, nonce, is_compressed = decrypt_data(b64decode(item["data"]), key)
-            if data is not None:
-                try:
-                    data = json.loads(data)
-                    item["data"] = data
-                    item["nonce"] = nonce
-                    item["decrypted"] = True
-                    item["compressed"] = is_compressed
-                    continue
+        key = keystore[meta]
 
-                except json.JSONDecodeError:
-                    pass
+        data, nonce, is_compressed = decrypt_data(b64decode(item["data"]), key)
+        if data is not None:
+            try:
+                data = json.loads(data)
+                item["data"] = data
+                item["nonce"] = nonce
+                item["decrypted"] = True
+                item["compressed"] = is_compressed
+                continue
+
+            except json.JSONDecodeError:
+                pass
 
         item["decrypted"] = False
 
@@ -263,38 +246,45 @@ def encrypt_user_data(items: Union[List[Dict[str, Any]], Dict[str, Any]], keysto
     if not isinstance(items, (list, tuple)):
         items = [items]
 
-    for item in items:
-        assert isinstance(item, dict), f"Expected type 'dict' for item of 'items', got type '{type(item)}'"
+    for i, item in enumerate(items):
+        if not isinstance(item, dict):
+            raise ValueError(f"Expected type 'dict' for item #{i} of 'items', got type '{type(item)}'")
 
         if "decrypted" in item:
             if item["decrypted"]:
-                # FIXME: replace the assert by meaningful errors ? Expect the data to be right ?
-                assert "data" in item, "Expected key 'data' in item"
-                assert "meta" in item, "Expected key 'meta' in item"
-                assert "nonce" in item, "Expected key 'nonce' in item"
-                assert "compressed" in item, "Expected key 'compressed' in item"
+                if "data" not in item:
+                    raise ValueError(f"Expected key 'data' in item #{i} of 'items'")
+                if "meta" not in item:
+                    raise ValueError(f"Expected key 'meta' in item #{i} of 'items'")
+                if "nonce" not in item:
+                    raise ValueError(f"Expected key 'nonce' in item #{i} of 'items'")
+                if "compressed" not in item:
+                    raise ValueError(f"Expected key 'compressed' in item #{i} of 'items'")
 
                 meta = item["meta"]
-                #            assert meta in keystore["keys"]
                 if meta not in keystore:
-                    print("Meta missing:", meta)
-                else:
-                    key = keystore[meta]
+                    raise NovelAIError("<UNKNOWN>", -1, f"Meta of item #{i} ({meta}) missing from keystore")
 
-                    data = json.dumps(item["data"], separators=(",", ":"), ensure_ascii=False)
-                    data = b64encode(encrypt_data(data, key, item["nonce"], item["compressed"])).decode()
+                key = keystore[meta]
 
-                    item["data"] = data
-                    del item["nonce"]
-                    del item["compressed"]
+                data = json.dumps(item["data"], separators=(",", ":"), ensure_ascii=False)
+                data = b64encode(encrypt_data(data, key, item["nonce"], item["compressed"])).decode()
+
+                item["data"] = data
+                del item["nonce"]
+                del item["compressed"]
 
             del item["decrypted"]
 
 
-def assign_content_to_story(
+def link_content_to_story(
     stories: Dict[str, Union[str, int, Dict[str, Any]]],
     story_contents: Union[List[Dict[str, Any]], Dict[str, Any]],
 ):
+    """
+    Link the story content to each story in :ref: stories
+    """
+
     if not isinstance(stories, (list, tuple)):
         stories = [stories]
 
@@ -311,27 +301,50 @@ def assign_content_to_story(
                 story["content"] = story_contents[remote_id]
 
 
-def remove_non_decrypted_user_data(items: List[Dict[str, Any]]):
-    i = 0
+def unlink_content_from_story(stories: Dict[str, Union[str, int, Dict[str, Any]]]):
+    """
+    Remove the story content from each story in :ref: stories
+    """
 
-    while i < len(items):
-        if items[i].get("decrypted", False) is False:
-            items.pop(i)
-        else:
-            i += 1
+    if not isinstance(stories, (list, tuple)):
+        stories = [stories]
+
+    for story in stories:
+        if story.get("decrypted") and "content" in story:
+            story.pop("content")
+
+
+def get_decrypted_user_data(items: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
+    """
+    Filter out items that have not been decrypted
+    """
+
+    return [item for item in items if item.get("decrypted", False)]
 
 
 def tokens_to_b64(tokens: Iterable[int]) -> str:
+    """
+    Encode a list of tokens into a base64 string that can be sent to the API
+    """
+
     return b64encode(b"".join(t.to_bytes(2, "little") for t in tokens)).decode()
 
 
 def b64_to_tokens(b64: str) -> List[int]:
+    """
+    Decode a base64 string returned by the API into a list of tokens
+    """
+
     b = b64decode(b64)
 
     return [int.from_bytes(b[i : i + 2], "little") for i in range(0, len(b), 2)]
 
 
 def extract_preset_data(presets: List[Dict[str, Any]]) -> Dict[str, Preset]:
+    """
+    Transform a list of preset data into a dict of Preset objects indexed by their id
+    """
+
     preset_list = {}
     for preset_data in presets:
         decompress_user_data(preset_data)
@@ -341,10 +354,16 @@ def extract_preset_data(presets: List[Dict[str, Any]]) -> Dict[str, Preset]:
 
 
 def tokenize_if_not(model: Model, o: Union[str, List[int]]) -> List[int]:
+    """
+    Tokenize the string if it is not already tokenized
+    """
+
     if isinstance(o, list):
         return o
 
-    assert isinstance(o, str)
+    if not isinstance(o, str):
+        raise ValueError(f"Expected type 'str' for 'o', got type '{type(o)}'")
+
     return Tokenizer.encode(model, o)
 
 
